/*
 * Copyright (c) 2008,2011 Intel Corporation.  All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the
 * GNU General Public License (GPL) Version 2, available from the
 * file COPYING in the main directory of this source tree, or the
 * OpenFabrics.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *	copyright notice, this list of conditions and the following
 *	disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <infiniband/driver.h>

#define SCIF_UVERBS_ABI_VERSION 6

int scif_query_device(struct ibv_context *context, struct ibv_device_attr *attr)
{
	unsigned major, minor, sub_minor;
	struct ibv_query_device	cmd;
	uint64_t raw_fw_ver;
	int ret;

	ret = ibv_cmd_query_device(context, attr, &raw_fw_ver, &cmd,
				   sizeof cmd);
	if (ret)
		return ret;

	major     = (raw_fw_ver >> 32) & 0xffff;
	minor     = (raw_fw_ver >> 16) & 0xffff;
	sub_minor = raw_fw_ver & 0xffff;

	snprintf(attr->fw_ver, sizeof attr->fw_ver, "%d.%d.%d",
		 major, minor, sub_minor);

	return 0;
}

int scif_query_port(struct ibv_context *context, uint8_t port,
		    struct ibv_port_attr *attr)
{
	struct ibv_query_port cmd;

	return ibv_cmd_query_port(context, port, attr, &cmd, sizeof cmd);
}

struct ibv_pd *scif_alloc_pd(struct ibv_context *context)
{
	struct ibv_pd *pd;
	struct ibv_alloc_pd cmd;
	struct ibv_alloc_pd_resp resp;

	pd = calloc(1, sizeof *pd);
	if (!pd)
		return NULL;

	if (ibv_cmd_alloc_pd(context, pd, &cmd, sizeof cmd, &resp,
			     sizeof resp)) {
		free(pd);
		return NULL;
	}

	return pd;
}

int scif_free_pd(struct ibv_pd *pd)
{
	int ret;

	ret = ibv_cmd_dealloc_pd(pd);
	if (!ret)
		free(pd);

	return ret;
}

struct ibv_qp *scif_create_qp(struct ibv_pd *pd, struct ibv_qp_init_attr *attr)
{
	struct ibv_qp *qp;
	struct ibv_create_qp cmd;
	struct ibv_create_qp_resp resp;
	int ret;

	qp = calloc(1, sizeof *qp);
	if (!qp)
		return NULL;

	ret = ibv_cmd_create_qp(pd, qp, attr, &cmd, sizeof cmd, &resp,
				sizeof resp);
	if (ret) {
		free(qp);
		return NULL;
	}

	return qp;
}

int scif_query_qp(struct ibv_qp *qp, struct ibv_qp_attr *attr,
		   int attr_mask, struct ibv_qp_init_attr *init_attr)
{
	struct ibv_query_qp cmd;

	return ibv_cmd_query_qp(qp, attr, attr_mask, init_attr,	&cmd,
				sizeof cmd);
}

int scif_modify_qp(struct ibv_qp *qp, struct ibv_qp_attr *attr, int attr_mask)
{
	struct ibv_modify_qp cmd;

	return ibv_cmd_modify_qp(qp, attr, attr_mask, &cmd, sizeof cmd);
}

int scif_destroy_qp(struct ibv_qp *ibqp)
{
	int ret;

	ret = ibv_cmd_destroy_qp(ibqp);
	if (!ret)
		free(ibqp);

	return ret;
}

struct ibv_cq *scif_create_cq(struct ibv_context *context, int cqe,
			      struct ibv_comp_channel *channel,
			      int comp_vector)
{
	struct ibv_cq *cq;
	struct ibv_create_cq cmd;
	struct ibv_create_cq_resp resp;
	int ret;

	cq = calloc(1, sizeof *cq);
	if (!cq)
		return NULL;

	ret = ibv_cmd_create_cq(context, cqe, channel, comp_vector,
				cq, &cmd, sizeof cmd, &resp, sizeof resp);
	if (ret) {
		free(cq);
		return NULL;
	}

	return cq;
}

int scif_resize_cq(struct ibv_cq *ibcq, int cqe)
{
	struct ibv_resize_cq cmd;
	struct ibv_resize_cq_resp resp;

	return ibv_cmd_resize_cq(ibcq, cqe, &cmd, sizeof cmd, &resp,
				 sizeof resp);
}

int scif_destroy_cq(struct ibv_cq *ibcq)
{
	int ret;

	ret = ibv_cmd_destroy_cq(ibcq);
	if (!ret)
		free(ibcq);

	return ret;
}

struct ibv_mr *scif_reg_mr(struct ibv_pd *pd, void *addr,
			   size_t length, int access)
{
	struct ibv_mr *mr;
	struct ibv_reg_mr cmd;
	struct ibv_reg_mr_resp resp;
	int ret;

	mr = calloc(1, sizeof *mr);
	if (!mr)
		return NULL;

	ret = ibv_cmd_reg_mr(pd, addr, length, (uintptr_t)addr, access, mr,
			     &cmd, sizeof cmd, &resp, sizeof(resp));
	if (ret) {
		free(mr);
		return NULL;
	}

	return mr;
}

int scif_dereg_mr(struct ibv_mr *mr)
{
	int ret;

	ret = ibv_cmd_dereg_mr(mr);
	if (!ret)
		free(mr);

	return ret;
}

struct ibv_ah *scif_create_ah(struct ibv_pd *pd, struct ibv_ah_attr *attr)
{
	struct ibv_ah *ah;
	int ret;

	ah = calloc(1, sizeof *ah);
	if (!ah)
		return NULL;

#ifdef HAVE_OLDER_VERBS
	ret = ibv_cmd_create_ah(pd, ah, attr);
#else
	struct ibv_create_ah_resp resp = { 0 };
	ret = ibv_cmd_create_ah(pd, ah, attr, &resp, sizeof(resp));

#endif

	if (ret) {
		free(ah);
		return NULL;
	}

	return ah;
}

int scif_destroy_ah(struct ibv_ah *ah)
{
	int ret;

	ret = ibv_cmd_destroy_ah(ah);
	if (!ret)
		free(ah);

	return ret;
}

struct ibv_srq *scif_create_srq(struct ibv_pd *pd,
				struct ibv_srq_init_attr *attr)
{
	struct ibv_srq *srq;
	struct ibv_create_srq cmd;
	struct ibv_create_srq_resp resp;
	int ret;

	srq = calloc(1, sizeof *srq);
	if (!srq)
		return NULL;

	ret = ibv_cmd_create_srq(pd, srq, attr, &cmd, sizeof cmd, &resp,
				 sizeof resp);
	if (ret) {
		free(srq);
		return NULL;
	}

	return srq;
}

int scif_modify_srq(struct ibv_srq *srq, struct ibv_srq_attr *attr,
		    int attr_mask)
{
	struct ibv_modify_srq cmd;

	return ibv_cmd_modify_srq(srq, attr, attr_mask, &cmd, sizeof cmd);
}

int scif_query_srq(struct ibv_srq *srq, struct ibv_srq_attr *attr)
{
	struct ibv_query_srq cmd;

	return ibv_cmd_query_srq(srq, attr, &cmd, sizeof cmd);
}

int scif_destroy_srq(struct ibv_srq *srq)
{
	int ret;

	ret = ibv_cmd_destroy_srq(srq);
	if (!ret)
		free(srq);

	return ret;
}

static struct ibv_context_ops scif_ctx_ops = {

	.query_device	= scif_query_device,
	.query_port	= scif_query_port,

	.alloc_pd	= scif_alloc_pd,
	.dealloc_pd	= scif_free_pd,

	.create_qp	= scif_create_qp,
	.query_qp	= scif_query_qp,
	.modify_qp	= scif_modify_qp,
	.destroy_qp	= scif_destroy_qp,

	.create_cq	= scif_create_cq,
	.resize_cq	= scif_resize_cq,
	.destroy_cq	= scif_destroy_cq,
	.poll_cq	= ibv_cmd_poll_cq,
	.req_notify_cq	= ibv_cmd_req_notify_cq,

	.reg_mr		= scif_reg_mr,
	.dereg_mr	= scif_dereg_mr,

	.post_send	= ibv_cmd_post_send,
	.post_recv	= ibv_cmd_post_recv,

	/* the following commands have not been implemented in the driver */
	.create_ah	= scif_create_ah,
	.destroy_ah	= scif_destroy_ah,
	.attach_mcast	= ibv_cmd_attach_mcast,
	.detach_mcast	= ibv_cmd_detach_mcast,

	.create_srq	= scif_create_srq,
	.modify_srq	= scif_modify_srq,
	.query_srq	= scif_query_srq,
	.destroy_srq	= scif_destroy_srq,

	.post_srq_recv	= ibv_cmd_post_srq_recv,
};

#ifdef HAVE_OLDER_VERBS
static struct ibv_context *scif_alloc_context(struct ibv_device *ibdev,
		int cmd_fd)
{
	struct ibv_context *context;
	struct ibv_get_context cmd;
	struct ibv_get_context_resp resp;

	context = calloc(1, sizeof *context);
	if (!context)
		return NULL;

	context->cmd_fd = cmd_fd;
	if (ibv_cmd_get_context(context, &cmd, sizeof cmd, &resp,
				sizeof resp)) {
		free(context);
		return NULL;
	}

	context->ops = scif_ctx_ops;
	return context;
}

static void scif_free_context(struct ibv_context *context)
{
	free(context);
}

static struct ibv_device_ops scif_dev_ops = {
	.alloc_context = scif_alloc_context,
	.free_context  = scif_free_context
};
#else
static int scif_init_context(struct verbs_device *dev, struct ibv_context *ctx,
		int cmd_fd)
{
	struct ibv_get_context cmd;
	struct ibv_get_context_resp resp;

	ctx->cmd_fd = cmd_fd;
	if (ibv_cmd_get_context(ctx, &cmd, sizeof cmd, &resp,
				sizeof resp)) {
		return -EINVAL;
	}

	ctx->ops = scif_ctx_ops;
	return 0;
}

static void scif_uninit_context(struct verbs_device *dev,
		struct ibv_context *context)
{
}

static struct verbs_device_ops scif_dev_ops = {
	.init_context = scif_init_context,
	.uninit_context  = scif_uninit_context
};
#endif

static struct ibv_device *scif_driver_init(const char *uverbs_sys_path,
		int abi_version)
{
	char value[16];
#ifdef HAVE_OLDER_VERBS
	struct ibv_device *dev;
#else
	struct verbs_device *dev;
#endif

#if 0
	int class;

	if (ibv_read_sysfs_file(uverbs_sys_path, "device/class", value,
				sizeof value) < 0)
		return NULL;

	class = strtol(value, NULL, 0);
	/* Look for Ethernet network class */
	if (class != (PCI_CLASS_NETWORK_ETHERNET << 8))
		return NULL;
#else
	/* This function is responsible for checking if the dev is expected */
	if (ibv_read_sysfs_file(uverbs_sys_path, "ibdev", value,
				sizeof value) < 0)
		return NULL;

	if (strncmp(value, "scif", 4))
		return NULL;
#endif

	if (abi_version > SCIF_UVERBS_ABI_VERSION) {
		fprintf(stderr, "Fatal: ABI version %d of %s"
				" is too new (expected %d)\n",
				abi_version, uverbs_sys_path, SCIF_UVERBS_ABI_VERSION);
		return NULL;
	}
	dev = calloc(1, sizeof *dev);
	if (!dev) {
		fprintf(stderr, "Fatal: couldn't allocate device for %s\n",
				uverbs_sys_path);
		return NULL;
	}

#ifdef HAVE_OLDER_VERBS
	dev->ops = scif_dev_ops;
#else
	dev->ops = &scif_dev_ops;
#endif

	return dev;
}

static __attribute__((constructor)) void scif_register_driver(void)
{
#ifdef HAVE_OLDER_VERBS
	ibv_register_driver("scif", scif_driver_init);
#else
	verbs_register_driver("scif", scif_driver_init);
#endif
}
